#!/usr/bin/env python3
"""Collection of custom types used in `pyrfp`."""
from typing import TypedDict

import torch
from pyapes.tools.spatial import Hess
from pyapes.tools.spatial import Jac
from pymytools.logger import Timer
from torch import Tensor


class Potentials(TypedDict):
    """Collection of potentials and their derivatives."""

    H: Tensor
    """Rosenbluth H potential."""
    jacH: Jac
    """Jacobian of the H potential."""
    G: Tensor
    """Rotational G potential."""
    jacG: Jac
    """Jacobian of the G potential."""
    hessG: Hess
    """Hessian of the H potential."""
    pdf: Tensor
    """Target distribution."""


class PotentialReturnType(TypedDict):
    pots: Potentials | None
    success: bool
    timer: Timer | None


class OptimizerConfig(TypedDict):
    lr: float
    max_iter: int | None
    gtol: float
    xtol: float


class LRSchedularOptions(TypedDict):
    t_max: int
    eta_min: float
    gamma: float | None
    verbose: bool
    start_after: int
    update_every: int


class LRSchedularConfig(TypedDict):
    name: str
    options: LRSchedularOptions


class LossOptions(TypedDict):
    weight: float
    factor: float
    update_every: int


class LossConfig(TypedDict):
    name: str
    options: LossOptions


class TrainingConfig(TypedDict):
    """Training configs"""

    data_dir: str
    """Training data directory"""
    model_dir: str
    """Model output directory"""
    n_epochs: int
    """Number of epochs"""
    layers: list[int]
    """Layer configuration (a number of neurons per layer). e.g. [10, 20, 30]"""
    act_funcs: list[str]
    """Activation function configuration (a name of activation functions per layer). e.g. ['relu', 'elu', 'relu']"""
    lr: float
    """Learning rate"""
    lr_schedular: LRSchedularConfig
    """Learning rate schedular config"""
    loss: LossConfig
    """Loss function"""
    dropout: float
    """Dropout rate"""
    n_batch: int
    """Number of batches"""
    batch_norm: bool
    """If true, perform batch normalization."""
    device: torch.device
    """It true, use GPU"""
    restart: str | None
    """If string is given, restart is the name of pre-trained model. Start training from that model. If None, start from the beginning."""
    data_chunk_size: int
    """The amount of data to be trained at once."""
    tracker_overwrite: bool
    """If true, Tensorboard will overwrite tracker."""
    num_workers: int
    """Number of CPUs for trainloader."""
    save_every: int
    """Save model at every `save_every` epochs."""
    dtype: torch.dtype
    """Data type for training."""



#!/usr/bin/env python
"""Tools for training models.
"""
import math
from typing import Callable
from typing import Optional

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.optim.lr_scheduler import CosineAnnealingLR

from pyrfp.types import LossConfig
from pyrfp.types import LRSchedularOptions


def set_act_funcs(act_f: list[str]) -> list[Callable]:
    """Setting activation function for each layer.
    sigmoid function is called from torch directly since F.sigmoid is
    deprecated.

    Supporting activation functions:
        relu, selu, elu, tanh, sigmoid

    Args:
        act_f (list[str]): list of activation functions

    Returns:
        list: list of activation functions
    """

    act_funcs = []

    for i_f in act_f:
        assert i_f in ACT_F_TYPES, f"Model: Unsupported activation function: {i_f}!"

        act_funcs.append(ACT_F_TYPES[i_f])

    return act_funcs


def diffRelu(input: Tensor) -> Tensor:
    """apply Relu activation function for diagonal diffusion matrix
    components

    This custom activation function is intended to ensure diagonal components
    of the diffusion matrix are always positive
    """
    diff_idx = [3, 5]

    result = input

    for idx in diff_idx:
        # use clone to avoid an inplace operation related error
        result[:, idx] = torch.relu(input[:, idx].clone())

    return result


def set_loss_funcs(loss_config: LossConfig) -> torch.nn.Module:
    """setting loss function used during the training"""

    # Make sure the lower case
    loss_f = loss_config["name"].lower()

    try:
        criterion = LOSS_TYPES[loss_f]
    except KeyError:
        raise KeyError(f"Unsupported loss function: {loss_f}!")

    if loss_f in LOSS_CUSTOM:
        criterion.set_options(loss_config["options"])

    return criterion


def set_lr_schedular(
    schedular: str,
    optimizer: torch.optim.Optimizer,
    opts: LRSchedularOptions | None = None,
) -> torch.optim.lr_scheduler._LRScheduler | None:
    """Setting learning rate schedular.

    Note:
        - Currently only available with ExponentialLR
    """
    from torch.optim.lr_scheduler import ExponentialLR

    if opts is None:
        lr_schedular = None
    else:
        verbose = opts["verbose"] if "verbose" in opts else False

        assert isinstance(verbose, bool)

        if schedular == "exp_lr":
            gamma = opts["gamma"]

            assert gamma is not None, "Model tools: gamma is not set!"

            lr_schedular = ExponentialLR(
                optimizer=optimizer, gamma=gamma, verbose=verbose
            )
        elif schedular == "cosine_lr":
            """CosineAnnealingLR"""
            t_max = int(opts["t_max"])
            eta_min = opts["eta_min"]

            lr_schedular = CosineAnnealingLR(
                optimizer=optimizer, T_max=t_max, eta_min=eta_min
            )

        elif schedular == None:
            lr_schedular = None
        else:
            msg = f"Unsupported schedular: {schedular}!"
            raise NameError(msg)

    return lr_schedular


class signedMSE(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, pred: Tensor, target: Tensor) -> Tensor:
        return _signed_MSELoss(pred, target)


class ALMSE(torch.nn.Module):
    """Asymmetric linear MSE loss function."""

    def __init__(self):
        super().__init__()

    def set_options(self, opts: Optional[dict] = None) -> None:
        """Update loss options."""

        if opts is None:
            self.n_slope = 5
        else:
            self.n_slope = opts["n_slope"]

    def forward(self, pred: Tensor, target: Tensor) -> Tensor:
        return _leakyRelu(pred, target, self.n_slope)

    def extra_repr(self) -> str:
        msg = f"n_slope: {self.n_slope}"

        return msg


class AEPMSE(torch.nn.Module):
    """Asymmetric energy preserving loss function."""

    def __init__(self):
        super().__init__()

    def set_options(self, opts: Optional[dict] = None) -> None:
        """Update loss options."""

        if opts is None:
            self.n_slope = 5
            self.weight = 0.01
        else:
            self.n_slope = opts["n_slope"]
            self.weight = opts["weight"]

    def forward(
        self,
        input: Tensor,
        pred: Tensor,
        target: Tensor,
        wiener: Tensor,
    ) -> Tensor:
        # Normal MSE
        loss_1 = _leakyRelu(pred, target, self.n_slope)
        loss_de_dt = ensemble_de_dt(input, pred, wiener)

        return loss_1 + loss_de_dt * self.weight

    def extra_repr(self) -> str:
        msg = f"  n_slope: {self.n_slope}\n" f"  weight: {self.weight}"

        return msg


class EPVIB(torch.nn.Module):
    """Energy preserving VIB loss function.
    Same energy preserving loss for the variational information bottleneck (VIB) model.
    """

    def __init__(self):
        super().__init__()

    def set_options(self, opts: Optional[dict] = None) -> None:
        """Update loss options."""

        if opts is None:
            self.weight_info = 0.01
            self.weight_de_dt = 0.01
        else:
            self.weight_info = opts["weight_info"]
            self.weight_de_dt = opts["weight_de_dt"]

    def extra_repr(self) -> str:
        msg = (
            f"  weight info: {self.weight_info}\n"
            f"  weight de_dt: {self.weight_de_dt}"
        )

        return msg

    def forward(
        self,
        input: Tensor,
        pred: Tensor,
        target: Tensor,
        wiener: Tensor,
        mu: Tensor,
        std: Tensor,
    ) -> Tensor:
        loss_mse = _mse_loss(pred, target)
        loss_info = -0.5 * (1 + 2 * std.log() - mu.pow(2) - std.pow(2)).sum(
            1
        ).mean().div(math.log(2))

        loss_de_dt = ensemble_de_dt(input, pred, wiener)

        return loss_mse + self.weight_info * loss_info + self.weight_de_dt * loss_de_dt


class EPMSE(torch.nn.Module):
    """Energy preserving MSE loss."""

    def __init__(self):
        super().__init__()

    def set_options(self, opts: Optional[dict] = None) -> None:
        """Update loss options."""

        if opts is None:
            self.weight = 0.01
        else:
            self.weight = opts["weight"]

    def forward(
        self,
        input: Tensor,
        pred: Tensor,
        target: Tensor,
        wiener: Tensor,
    ) -> Tensor:
        # MSE
        loss_mse = _mse_loss(pred, target)

        # de_dt
        loss_de_dt = ensemble_de_dt(input, pred, wiener)

        return loss_mse + self.weight * loss_de_dt

    def extra_repr(self) -> str:
        msg = f"weight: {self.weight}"

        return msg


class DEPMSE(torch.nn.Module):
    """Deterministic Energy preserving MSE loss."""

    def __init__(self):
        super().__init__()

    def forward(
        self,
        input: Tensor,
        pred: Tensor,
        target: Tensor,
    ) -> Tensor:
        loss_mse = _mse_loss(pred, target)
        loss_ep = ensemble_e(input, pred)
        loss_mp = ensemble_m(input, pred)

        return loss_mse + (loss_ep + loss_mp) * self.weight

    def set_options(self, opts: dict) -> None:
        """Update loss options."""

        self.weight = opts["weight"]
        self.factor = opts["factor"]
        self.update_every = opts["update_every"]

    def update_weight(self) -> None:
        """Update weight. Multiply by the factor."""

        self.weight *= self.factor

    def extra_repr(self) -> str:
        msg = (
            f"  weight: {self.weight}\n"
            f"  factor: {self.factor}\n"
            f"  weight update every: {self.update_every} epoch"
        )

        return msg


class DMEPMSE(torch.nn.Module):
    """Deterministic momentum and energy preserging MSE loss function."""

    def __init__(self):
        super().__init__()

    def forward(
        self,
        input: Tensor,
        pred: Tensor,
        target: Tensor,
    ) -> Tensor:
        loss_mse = _mse_loss(pred, target)
        loss_ep = ensemble_e(input, pred)
        loss_mp = ensemble_m(input, pred)

        return loss_mse + (loss_ep + loss_mp) * self.weight

    def set_options(self, opts: dict) -> None:
        """Update loss options."""

        self.weight = opts["weight"]
        self.factor = opts["factor"]
        self.update_every = opts["update_every"]

    def update_weight(self) -> None:
        """Update weight. Multiply by the factor."""

        self.weight *= self.factor

    def extra_repr(self) -> str:
        msg = (
            f"  weight: {self.weight}\n"
            f"  factor: {self.factor}\n"
            f"  weight update every: {self.update_every} epoch"
        )

        return msg


def ensemble_e(input: Tensor, pred: Tensor) -> Tensor:
    """Ensemble average of the kinetic energy. Exclude the effect from
    The random noise.
    """

    idx_set = [3, 4, 5, 4, 6, 7, 5, 7, 8]

    dt = 0.01
    dt2 = dt * dt

    uA = 2 * torch.sum(torch.mul(input[:, -3::], pred[:, :3]), dim=1) * dt

    A = torch.sum(torch.mul(pred[:, :3], pred[:, :3]), dim=1) * dt2

    D = torch.sum(torch.mul(pred[:, idx_set], pred[:, idx_set]), dim=1) * dt

    return torch.abs(uA + A + D).mean()


def _mse_loss(pred: Tensor, target: Tensor) -> Tensor:
    """Custom MSE loss. Nothing special."""

    return torch.mean((pred - target) ** 2)


def ensemble_m(input: Tensor, pred: Tensor):
    dt = 0.01

    uA = torch.sum(torch.mul(input[:, -3::], pred[:, :3]), dim=1) * dt

    # Return ensemble of de/dt
    return torch.abs(uA).mean()


def ensemble_de_dt(
    input: Tensor,
    pred: Tensor,
    wiener: Tensor,
) -> Tensor:
    r"""Calculate mean(de/dt).

    Which is calculated by

    .. math::

        \vert \frac{d e_{kin}}{dt} \vert
        = \vert u_{input} A_{pred}
        + u_{input}\sqrt{D_{pred}}\mathcal{N}(0, \frac{1}{dt}) \vert = 0 \\
        \text{where,} e = \frac{1}{2} u_{input}^2

    Note:
        - Here, \(\mathcal{N}(0, \frac{1}{dt})\) is
          pre-calculated and stored in the target files. -> target[:, -3::]

    """

    idx_set = [[3, 4, 5], [4, 6, 7], [5, 7, 8]]

    diff = torch.zeros_like(wiener)
    # pred 3, 4, 5, 6, 7, 8 => 11, 12, 13, 22, 23, 33
    for i, idx in enumerate(idx_set):
        diff[:, i] = torch.sum(torch.mul(pred[:, idx], wiener), dim=1)

    de_dt = torch.sum(torch.mul(input[:, -3::], pred[:, :3]), dim=1) + torch.sum(
        torch.mul(input[:, -3::], diff), dim=1
    )

    # Return ensemble of de/dt
    return torch.abs(de_dt.mean())


def _signed_MSELoss(pred: Tensor, target: Tensor) -> Tensor:
    """enforce prediction to have same sign with target."""

    # if pred.sign() and target.sign() are different, assign zero
    # then multiply target.sign() to make sure have same sign with the target
    signed_pred = F.relu(pred * target.sign()) * target.sign()

    return torch.mean((signed_pred - target) ** 2)


def _leakyRelu(pred: Tensor, target: Tensor, n_slope: float) -> Tensor:
    check_sign = pred * target.sign() < 0

    # assign strong gradient on negative side
    strong_leakyRelu = torch.nn.LeakyReLU(n_slope)

    pred[check_sign] = strong_leakyRelu(pred[check_sign])

    return torch.mean((pred - target) ** 2)


LOSS_TYPES = {
    "mse": torch.nn.MSELoss(),
    "l1": torch.nn.L1Loss(),
    "sl1": torch.nn.SmoothL1Loss(),
    "signedmse": signedMSE(),
    "almse": ALMSE(),
    "epmse": EPMSE(),
    "aepmse": AEPMSE(),
    "epvib": EPVIB(),
    "depmse": DEPMSE(),
    "dmepmse": DMEPMSE(),
}
"""Interface for the loss function."""

LOSS_CUSTOM = ["epmse", "aepmse", "depmse", "dmepmse"]
"""List of custom loss function."""

ACT_F_TYPES: dict[str, Callable] = {
    "relu": F.relu,
    "selu": F.selu,
    "elu": F.elu,
    "tanh": torch.tanh,
    "sigmoid": torch.sigmoid,
    "diffrelu": diffRelu,
}
"""Interface for the activation function."""



#!/usr/bin/env python3
import torch
from torch import Tensor

from pyrfp.model_tools import set_act_funcs


class NLinearLayerModel(torch.nn.Module):
    """Class for neural network (nn) model which consisted of N number of
    hidden layers.
    Therefore, total number of layer is, N + 2 including input layer and
    prediction layer.

    Note:
        I implemented batch normalization here, but no need to use
        (actually not working in our case) since our prediction process
        only requires 1D data.
    """

    def __init__(
        self,
        d_output: int = 12,
        d_input: int = 8,
        d_hidden: list[int] = [128, 128],
        act_func: list[str] = ["tanh", "tanh"],
        batch_norm: bool = False,
        dropout_rate: float = 0.0,
        dtype: torch.dtype = torch.float32,
    ):
        """Construct NLinearLayerModel.

        Args:
            torch (object): base of class
            d_output (int): data output size
            d_input (int): data input size
            d_hidden (list): number of hidden layers
                e.g.) [30, 40, 50, 20]
            act_func (list, optional): activation function.
                Defaults to relu.
                e.g.) ['relu', 'elu', 'sigmoid', 'relu', 'tanh']
            batch_norm (bool, optional): batch normalization.
            dropout_rate (float, optional): dropout rate.
        """
        super().__init__()

        self.n_hidden = len(d_hidden)

        self.act_func = act_func
        self.batch_norm = batch_norm
        self.hidden = []
        self.hidden_bn = []
        self.act_funcs = set_act_funcs(act_func)
        self.dropout_rate = dropout_rate

        self.layers = [d_input] + d_hidden

        # construct hidden layers
        for i_h in range(self.n_hidden):
            self.hidden.append(
                torch.nn.Linear(self.layers[i_h], self.layers[i_h + 1], dtype=dtype)
            )

        # convert to ModuleList. python list is not working here.
        self.hidden = torch.nn.ModuleList(self.hidden)

        if self.batch_norm:
            # including input batch normalization
            for i_bn in range(self.n_hidden + 1):
                self.hidden_bn.append(torch.nn.BatchNorm1d(self.layers[i_bn]))
            self.hidden_bn = torch.nn.ModuleList(self.hidden_bn)

        # prediction layer
        self.predict = torch.nn.Linear(d_hidden[-1], d_output, dtype=dtype)

        if self.dropout_rate > 0:
            # dropout
            self.dropout = torch.nn.Dropout(p=self.dropout_rate)

    def forward(self, x: Tensor):
        """Feedforward process"""

        if self.batch_norm:
            x = self.hidden_bn[0](x)

        for i_h in range(self.n_hidden):
            # activation function
            if self.batch_norm:
                x = self.act_funcs[i_h](self.hidden_bn[i_h + 1](self.hidden[i_h](x)))
            else:
                x = self.act_funcs[i_h](self.hidden[i_h](x))

            if self.dropout_rate > 0:
                x = self.dropout(x)

        if self.n_hidden == len(self.act_func):
            x = self.predict(x)
        elif self.n_hidden < len(self.act_func):
            # apply custom activation function for the prediction layer
            x = self.act_funcs[-1](self.predict(x))

        return x

    def extra_repr(self) -> str:
        """Custom repr"""
        msg = "(act_funcs): (\n"
        for i, func in enumerate(self.act_func):
            msg += f"  ({i}): {func}"
            if i + 1 > self.n_hidden:
                msg += " <- custom activation function for the (predict) layer"
            msg += "\n"
        msg += ")"

        return msg



#!/usr/bin/env python3
"""Training data generation for the Rosenbluth Fokker Planck equation.

Here, we compute the Rosenbluth potentials using the maximum entropy distribution and the Green's function (for the boundary conditions).

In this case, the computation of the boundary condition (due to the evaluation of the Greens' function) is the most time consuming part.

To have a solution of the Green's function, we can take several approaches:
    - Naive summation
    - Loop over vectorized sum
    - Reshape all tensors and use the vectorized sum
    - Use vmap

The performance of the abovementioned approaches are following:
    - For 32 x 64 grid
    ┏━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓
    ┃  Schemes   ┃ Elapsed time (s)     ┃
    ┡━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩
    │ Naive sum  │ 55.80528954183683    │
    │  Loop sum  │ 0.027860709000378847 │
    │Reshaped sum│ 0.03574174991808832  │
    │  Vmap sum  │ 0.009342916077002883 │
    └────────────┴──────────────────────┘

Note:
    - Above performance results are illusion. `vmap` works very well for the native torch functions but not for my `ellipe` and `ellipk` functions (from `pymytools`).
    - Therefore, we will use `Reshaped sum` approach for the computation of the potential boundary conditions.

Therefore, we decided to use the vmap approach for the computation of the boundary conditions.
"""
from dataclasses import dataclass
from math import pi

import pymaxed
import torch
from pyapes.core.mesh import Mesh
from pyapes.core.solver.fdm import FDM
from pyapes.core.solver.ops import Solver
from pyapes.core.solver.tools import FDMSolverConfig
from pyapes.core.variables import Field
from pyapes.core.variables.bcs import CylinderBoundary
from pyapes.tools.spatial import ScalarOP
from pymaxed.maxed import Maxed
from pymaxed.vectors import Vec
from pymytools.logger import logging
from pymytools.logger import markup
from pymytools.logger import Timer
from pymytools.special import ellipe
from pymytools.special import ellipk
from scipy.special import ellipe as s_ellipe
from scipy.special import ellipk as s_ellipk
from torch import Tensor
from torch import vmap

from pyrfp import __version__
from pyrfp.types import OptimizerConfig
from pyrfp.types import PotentialReturnType


@dataclass
class RosenbluthPotentials_RZ:
    """Compute the Rosenbluth potentials.

    Note:
        - This object is only designed to be worked in the `rz` coordinate for the moment order up to 4th order.
        - More Generalized case will be a future plan.

    Args:
        mesh (Mesh): Mesh object.
    """

    mesh: Mesh
    optimizer_config: OptimizerConfig | None = None
    solver_config: FDMSolverConfig | None = None

    def __post_init__(self):
        """Post initialization."""

        assert (
            self.mesh.domain.type == "cylinder"
        ), "RosenbluthPotential_RZ only works for cylinder domain."

        self.timer = Timer()

    def from_moments(self, mnts: Tensor) -> PotentialReturnType:
        """From moment, obtain the maximum entropy distribution (maxed)"""

        assert (
            self.optimizer_config is not None
        ), "RosenbluthPotential_rz: optimizer config is not provided."

        assert (
            self.solver_config is not None
        ), "RosenbluthPotential_rz: solver config is not provided."

        logging.info(
            markup(
                f"🚀 Computing Maximum Entropy Distribution  🚀",
                "red",
                "italic",
            )
            + f" (pymaxed v {pymaxed.__version__})"
        )
        logging.info("Target moment: " + markup(f" {mnts}", "blue"))
        logging.info(
            "Domain: "
            + markup(
                f" {self.mesh.lower.tolist()} x {self.mesh.upper.tolist()}", "blue"
            )
        )
        logging.info("Grid: " + markup(f" {list(self.mesh.nx)}", "blue"))

        vec = Vec(self.mesh, mnts, 4, [50, 100])
        maxed = Maxed(
            vec,
            lr=self.optimizer_config["lr"],
            max_itr=self.optimizer_config["max_iter"],
            gtol=self.optimizer_config["gtol"],
            xtol=self.optimizer_config["xtol"],
            disp=False,
        )

        logging.info(markup("Solving for MaxEd...", "yellow"))
        self.timer.start("maxed")
        maxed.solve()
        self.timer.end("maxed")
        logging.info(
            "🔥Done in " + markup(f"{self.timer.elapsed('maxed'):.2f} s", "blue")
        )
        if not maxed.success:
            return {"pots": None, "success": False, "timer": None}

        pdf = maxed.dist_from_coeffs()

        # Density correction
        dr = self.mesh.grid[0][1, 0] - self.mesh.grid[0][0, 0]
        dz = self.mesh.grid[1][0, 1] - self.mesh.grid[1][0, 0]

        density = torch.sum(2.0 * torch.pi * self.mesh.grid[0] * pdf) * dr * dz
        pdf /= density

        return self.from_pdf(pdf)

    def from_pdf(self, pdf: Tensor) -> PotentialReturnType:
        """Compute the Rosenbluth potentials and their derivatives (Jacobian and Hessian) from the given PDF by solving the Poisson equation using the iterative solver (bicgstab scheme).

        Args:
            pdf (Tensor): a given PDF tensor.

        Returns:
            PotentialReturnType: a dictionary containing the potentials and their derivatives.

        Note:
            - The components of the Jacobian and Hessian can be accessed by member attribute of each object, e.g. `PotentialReturnType["jacH"].r` for the radial component of the Jacobian.
            - The `PotentialReturnType["success"]` is `True` only H and G potential both are converged.
        """

        assert (
            self.solver_config is not None
        ), "RosenbluthPotential_rz: solver config is not provided."

        solver = Solver({"fdm": self.solver_config})
        fdm = FDM()

        # Our Scalar field has additional dimension other than the mesh dimension in the leading position.
        # Therefore, if pdf is just like mesh dimension, add one as for the leading dimension
        if pdf.shape[0] != 1 or pdf.shape == self.mesh.nx:
            pdf = pdf.unsqueeze(0)

        logging.info(
            markup(
                f"🚀 Computing Rosenbluth potentials 🚀",
                "red",
                "italic",
            )
            + f" (pyrfp v {__version__})"
        )
        logging.info(
            "Domain: "
            + markup(
                f" {self.mesh.lower.tolist()} x {self.mesh.upper.tolist()}", "blue"
            )
        )
        logging.info("Grid: " + markup(f" {list(self.mesh.nx)}", "blue"))
        logging.info(markup("Evaluating H potential boundary...", "yellow"))
        self.timer.start("H_bc")
        bc_vals = get_analytic_bcs(self.mesh, pdf[0], "H")
        self.timer.end("H_bc")
        logging.info(
            "🔥Done in " + markup(f"{self.timer.elapsed('H_bc'):.2f} s", "blue")
        )

        bc_H = _set_bc_rz(bc_vals)
        H_pot = Field("H", 1, self.mesh, {"domain": bc_H(), "obstacle": None})

        logging.info(markup("Solving H potential...", "yellow"))
        self.timer.start("H_pot")
        solver.set_eq(fdm.laplacian(H_pot) == -8 * pi * pdf)
        solver.solve()
        self.timer.end("H_pot")
        logging.info(
            "🔥Done in " + markup(f"{self.timer.elapsed('H_pot'):.2f} s", "blue")
        )
        h_success = solver.report["converge"]

        logging.info(markup("Evaluating G potential boundary...", "yellow"))
        self.timer.start("G_bc")
        bc_vals = get_analytic_bcs(self.mesh, pdf[0], "G")
        self.timer.end("G_bc")
        logging.info(
            "🔥Done in " + markup(f"{self.timer.elapsed('G_bc'):.2f} s", "blue")
        )

        bc_G = _set_bc_rz(bc_vals)
        G_pot = Field("G", 1, self.mesh, {"domain": bc_G(), "obstacle": None})

        logging.info(markup("Solving G potential...", "yellow"))
        self.timer.start("G_pot")
        solver.set_eq(fdm.laplacian(G_pot) == H_pot())
        solver.solve()
        self.timer.end("G_pot")
        logging.info(
            "🔥Done in " + markup(f"{self.timer.elapsed('G_pot'):.2f} s", "blue")
        )
        g_success = solver.report["converge"]

        logging.info(
            "H: "
            + markup(
                f"{'successful' if h_success else 'fail'}",
                "green" if h_success else "red",
            )
            + ", G: "
            + markup(
                f"{'successful' if g_success else 'fail'}",
                "green" if g_success else "red",
            )
        )
        logging.info(markup("🎉 Finish! 🎉 ", "red"))

        return {
            "pots": {
                "H": H_pot()[0],
                "jacH": ScalarOP.jac(H_pot),
                "G": G_pot()[0],
                "jacG": ScalarOP.jac(G_pot),
                "hessG": ScalarOP.hess(G_pot),
                "pdf": pdf[0],
            },
            "success": h_success & g_success,
            "timer": self.timer,
        }


def get_analytic_bcs(
    mesh: Mesh, pdf: Tensor, pot: str, cpu: bool = True
) -> dict[str, Tensor]:
    bcs: dict[str, Tensor] = {}

    for k, v in mesh.d_mask.items():
        if k == "rl":
            pass
        if cpu:
            bcs[k] = analytic_potentials_rz_cpu(
                (mesh.grid[0][v], mesh.grid[1][v]), mesh.grid, pdf, pot
            )

        else:
            bcs[k] = analytic_potentials_rz(
                (mesh.grid[0][v], mesh.grid[1][v]), mesh.grid, pdf, pot
            )

    return bcs


def _set_bc_rz(vals: dict[str, Tensor]) -> CylinderBoundary:
    return CylinderBoundary(
        rl={"bc_type": "neumann", "bc_val": 0.0},
        ru={"bc_type": "dirichlet", "bc_val": vals["ru"]},
        zl={"bc_type": "dirichlet", "bc_val": vals["zl"]},
        zu={"bc_type": "dirichlet", "bc_val": vals["zu"]},
    )


def analytic_potentials_rz_cpu(
    target: tuple[Tensor, ...], grid: tuple[Tensor, ...], pdf: Tensor, potential: str
) -> Tensor:
    """Since vectorized version is too slow, lets try CPU only version using `scipy.special`

    Note:
        - Literal translation of the vectorized version, `analytic_potentials_rz`
    """
    ur = grid[0]
    uz = grid[1]

    hr = ur[1, 0] - ur[0, 0]
    hz = uz[0, 1] - uz[0, 0]

    # What if target[0].shape = torch.Size([64]) and grid[0].shape = torch.Size([64, 128?
    # I need to make 64 x 1 x (64 * 128)
    ur_target = target[0].flatten().unsqueeze(1).repeat(1, ur.numel())
    uz_target = target[1].flatten().unsqueeze(1).repeat(1, ur.numel())

    ur = ur.flatten().repeat(target[0].numel(), 1)
    uz = uz.flatten().repeat(target[0].numel(), 1)
    pdf = pdf.flatten().repeat(target[0].numel(), 1)

    inner = (ur_target + ur) ** 2 + (uz_target - uz) ** 2
    k = 4 * ur_target * ur / inner

    if potential.lower() == "h":
        ek = s_ellipk(k.to(device=torch.device("cpu"))).to(
            device=ur.device, dtype=ur.dtype
        )
        ek[k.eq(1.0)] = 0.0
        return torch.sum(
            torch.nan_to_num(
                8 * ur * pdf * ek / torch.sqrt(inner) * hr * hz,
                nan=0.0,
                posinf=0.0,
                neginf=0.0,
            ),
            dim=1,
        ).view(*target[0].shape)
    elif potential.lower() == "g":
        ee = s_ellipe(k.to(device=torch.device("cpu"))).to(
            device=ur.device, dtype=ur.dtype
        )
        ee[k.eq(1.0)] = 0.0
        return torch.sum(
            torch.nan_to_num(
                4 * ur * pdf * ee * torch.sqrt(inner) * hr * hz,
                nan=0.0,
                posinf=0.0,
                neginf=0.0,
            ),
            dim=1,
        ).view(*target[0].shape)
    else:
        raise ValueError("Potential must be either H or G")


def analytic_potentials_rz(
    target: tuple[Tensor, ...], grid: tuple[Tensor, ...], pdf: Tensor, potential: str
) -> Tensor:
    r"""Compute analytic solution of the Rosenbluth potential using the Green's formulation in the axisymmetric phase space.
    The integration is done with the elliptic integral of the first and second kind.

    The H potential is computed by:

    .. math::

        \left. H(u_\perp, u_\parallel) \right|_{\Omega_{BC}} = 8\int_{0}^{u_\perp^{max}}
        \int_{-u_\parallel^{max}}^{u_\parallel^{max}}
        u_\perp \frac{f^{MED}_{\vec{\lambda}}(u_\perp', u_\parallel')K[k(u_\perp, u_\parallel;u_\perp', u_\parallel')]}
        {\sqrt{(u_\perp + u_\perp')^2 + (u_\parallel - u_\parallel')^2}}du_\perp'du_\parallel' \quad \text{and}

    And the G potential is:

    .. math::

        \left. G(u_\perp, u_\parallel) \right|_{\Omega_{BC}} &=& 4\int_{0}^{u_\perp^{max}}
        \int_{-u_\parallel^{max}}^{u_\parallel^{max}}
        u_\perp f^{MED}_{\vec{\lambda}}(u_\perp', u_\parallel')E[k(u_\perp, u_\parallel;u_\perp', u_\parallel')] \nonumber\\
        &&\times \sqrt{(u_\perp + u_\perp')^2 + (u_\parallel - u_\parallel')^2}du_\perp'du_\parallel'

    Examples:

        >>> r = torch.linspace(0, 1, 32)
        >>> z = torch.linspace(-1, 1, 64)
        >>> grid = torch.meshgrid(r, z, indexing="ij")
        >>> pdf = torch.tensor(...)
        >>> H_pot = analytic_potentials_rz(grid, grid, pdf, "H") # H potential in the whole domain, target == grid
        >>> G_pot = analytic_potentials_rz(grid, grid, pdf, "G") # G potential in the whole domain, target == grid

    Args:
        target (tuple[Tensor, ...]): The target phase space to evaluate the potentials.
        grid (tuple[Tensor, ...]): The grid of the phase space.
        pdf (Tensor): The maximum entropy distribution.
        potential (str): The potential to compute. Either `H` or `G`.

    Returns:
        Tensor: The potential evaluated at the grid in the domain.

    Note:
        - The target and grid are separated and `target` $\in$ `grid`. This is intended to utilize this function for the evaluation of the boundary condition.
        - If you want to solve the potential in the whole domain, you can use the `target` and `grid` as the same.
    """

    ur = grid[0]
    uz = grid[1]

    hr = ur[1, 0] - ur[0, 0]
    hz = uz[0, 1] - uz[0, 0]

    if potential.lower() == "h":
        # Compute H potential
        vmap_func = vmap(
            lambda x, y: torch.sum(
                torch.nan_to_num(
                    ellipk(
                        4 * x * ur / ((x + ur) ** 2 + (y - uz) ** 2),
                    )
                    * 8
                    * ur
                    * pdf
                    / (torch.sqrt((x + ur) ** 2 + (y - uz) ** 2))
                    * hr
                    * hz,
                    nan=0.0,
                    posinf=0.0,
                    neginf=0.0,
                )
            )
        )

        return vmap_func(target[0].flatten(), target[1].flatten()).view(
            *target[0].shape
        )

    elif potential.lower() == "g":
        # Compute G potential
        vmap_func = vmap(
            lambda x, y: torch.sum(
                torch.nan_to_num(
                    ellipe(
                        4 * x * ur / ((x + ur) ** 2 + (y - uz) ** 2),
                    )
                    * 4
                    * ur
                    * pdf
                    * (torch.sqrt((x + ur) ** 2 + (y - uz) ** 2))
                    * hr
                    * hz,
                    nan=0.0,
                    posinf=0.0,
                    neginf=0.0,
                )
            )
        )

        return vmap_func(target[0].flatten(), target[1].flatten()).view(
            *target[0].shape
        )

    else:
        raise ValueError("Potential must be either 'H' or 'G'.")
